﻿using FluentAssertions;
using MongoDB.Driver;
using NSubstitute;
using Robware.Blog;
using Robware.Blog.Data.State;
using System;
using System.Threading;
using System.Threading.Tasks;
using Xunit;

namespace Robware.Data.Tests {
	public class BlogRepositoryTests {
		[Fact]
		public async Task SavePost_WithNewPost_GetsNewIdAndSetsTimestampAndSavesAsync() {
			var findResult = Substitute.For<IAsyncCursor<BlogPostState>>();
			findResult.Current.Returns(new[] { new BlogPostState { Id = 10 } });
			findResult.MoveNext(Arg.Any<CancellationToken>()).Returns(true, false);
			findResult.MoveNextAsync(Arg.Any<CancellationToken>()).Returns(true, false);
			var collection = Substitute.For<IMongoCollection<BlogPostState>>();
			collection.FindAsync(Arg.Any<FilterDefinition<BlogPostState>>(), Arg.Any<FindOptions<BlogPostState>>()).Returns(findResult);
			var database = Substitute.For<IMongoDatabase>();
			database.GetCollection<BlogPostState>("blog").Returns(collection);

			var repo = new BlogRepository(database);
			var post = new BlogPost();
			await repo.SavePost(post);
			post.Id.Should().Be(11);
			post.Timestamp.Should().BeCloseTo(DateTime.Now);
			await collection.Received(1).ReplaceOneAsync(Arg.Any<FilterDefinition<BlogPostState>>(), Arg.Any<BlogPostState>(), Arg.Any<ReplaceOptions>());
		}

		[Fact]
		public async Task SavePost_WithExistingPost_DoesNotSetTimestampAndSavesAsync() {
			var collection = Substitute.For<IMongoCollection<BlogPostState>>();
			var database = Substitute.For<IMongoDatabase>();
			database.GetCollection<BlogPostState>("blog").Returns(collection);

			var repo = new BlogRepository(database);
			var post = new BlogPost() {
				Id = 1
			};
			await repo.SavePost(post);
			post.Id.Should().Be(1);
			post.Timestamp.Should().Be(new DateTime());
			await collection.Received(1).ReplaceOneAsync(Arg.Any<FilterDefinition<BlogPostState>>(), Arg.Any<BlogPostState>(), Arg.Any<ReplaceOptions>());
		}

		[Fact]
		public async Task GetCountAsync_ReturnsCountFromDatabase() {
			var collection = Substitute.For<IMongoCollection<BlogPostState>>();
			collection.CountDocumentsAsync(Arg.Any<FilterDefinition<BlogPostState>>()).Returns(10);
			var database = Substitute.For<IMongoDatabase>();
			database.GetCollection<BlogPostState>("blog").Returns(collection);

			var repo = new BlogRepository(database);
			(await repo.GetCountAsync()).Should().Be(10);
		}
	}
}
