﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Threading.Tasks;
using MongoDB.Driver;
using Robware.Blog;
using Robware.Blog.Data.State;

namespace Robware.Data {
	public class BlogRepository : IBlogRepository {
		private readonly IMongoCollection<BlogPostState> _collection;

		public BlogRepository(IMongoDatabase database) {
			_collection = database.GetCollection<BlogPostState>("blog");
		}

		private BlogPost MapStateToPost(BlogPostState state) =>
			new BlogPost {
				Id = state.Id,
				Title = state.Title,
				Content = state.Content,
				Timestamp = state.Timestamp,
				Draft = state.Draft,
				Url = state.Url,
				UserId = state.UserId,
			};

		private IEnumerable<BlogPost> MapStateToPost(IEnumerable<BlogPostState> states) => states.Select(MapStateToPost);

		public async Task DeletePostAsync(int id) {
			var update = Builders<BlogPostState>.Update.Set(nameof(BlogPostState.Deleted), true);
			await _collection.UpdateOneAsync(post => post.Id == id, update);
		}

		public async Task<IEnumerable<BlogPost>> GetAllPostsAsync() => MapStateToPost((await _collection.FindAsync(post => !post.Deleted)).ToEnumerable());

		private async Task<int> GetCountAsync(bool includeAll) => (int)await _collection.CountDocumentsAsync(post => includeAll || (!post.Deleted && string.IsNullOrEmpty(post.Content)));

		public async Task<int> GetCountAsync() => await GetCountAsync(false);

		public async Task<BlogPost> GetLatestPostAsync() => (await GetLatestPostsAsync(1)).FirstOrDefault();

		public async Task<IEnumerable<BlogPost>> GetLatestPostsAsync(int limit, int offset = 0) {
			var filter = Builders<BlogPostState>.Filter.And(
				Builders<BlogPostState>.Filter.Eq(post => post.Deleted, false),
				Builders<BlogPostState>.Filter.Ne(post => post.Content, null)
			);
			var sort = Builders<BlogPostState>.Sort.Descending(post => post.Timestamp);
			var options = new FindOptions<BlogPostState> {
				Sort = sort,
				Limit = limit,
				Skip = offset
			};
			var states = (await _collection.FindAsync<BlogPostState>(filter, options)).ToEnumerable();
			return MapStateToPost(states);
		}

		private async Task<BlogPost> GetPostAsync(Expression<Func<BlogPostState, bool>> filter) {
			var result = (await _collection.FindAsync(filter)).FirstOrDefault();
			if (result == null)
				throw new ItemNotFoundException(nameof(GetPostAsync), null);
			return MapStateToPost(result);
		}

		public async Task<BlogPost> GetPostByIdAsync(int id) => await GetPostAsync(post => post.Id == id && !post.Deleted);

		public async Task<BlogPost> GetPostByUrlAsync(string url) => await GetPostAsync(post => post.Url == url && !post.Deleted);

		private async Task<int> GetLastPostId() {
			var sort = Builders<BlogPostState>.Sort.Descending(post => post.Id);
			var options = new FindOptions<BlogPostState> {
				Sort = sort,
				Limit = 1
			};
			var states = (await _collection.FindAsync<BlogPostState>(_ => true, options)).ToEnumerable();
			return states.First().Id;
		}

		public async Task<BlogPost> SavePost(BlogPost post) {
			if (post.Id == 0) {
				post.Id = (await GetLastPostId()) + 1;
				post.Timestamp = DateTime.Now;
			}

			var mongoPost = new BlogPostState(post);
			var filter = Builders<BlogPostState>.Filter.Eq(p => p.Id, post.Id);
			var options = new ReplaceOptions {
				IsUpsert = true
			};
			await _collection.ReplaceOneAsync(filter, mongoPost, options);

			return post;
		}
	}
}
